@@ -1,15 +1,12 @@
-/* Copyright (C) 1994, Memorial University of Newfoundland.
- * This file is covered by the GNU General Public License, version 2, see
- * the file misc/COPYING for details.
- *
- * acconfig.h for the PD ksh
+/*
+ * This file, acconfig.h, which is a part of pdksh (the public domain ksh),
+ * is placed in the public domain.  It comes with no licence, warranty
+ * or guarantee of any kind (i.e., at your own risk).
  */
 
 #ifndef CONFIG_H
 #define CONFIG_H
 
-#include "options.h"
-
 @TOP@
 
 /* Define if your kernal doesn't handle scripts starting with #! */
@@ -18,6 +15,9 @@
 /* Define if dup2() preserves the close-on-exec flag (ultrix does this) */
 #undef DUP2_BROKEN
 
+/* Define as the return value of signal handlers (0 or ).  */
+#undef RETSIGVAL
+
 /* Define if you have posix signal routines (sigaction(), et. al.) */
 #undef POSIX_SIGNALS
 
@@ -33,6 +33,9 @@
 /* Define to use the fake posix signal routines (sigact.[ch]) */
 #undef USE_FAKE_SIGACT
 
+/* Define if signals don't interrupt read() */
+#undef SIGNALS_DONT_INTERRUPT
+
 /* Define if you have bsd versions of the setpgrp() and getpgrp() routines */
 #undef BSD_PGRP
 
@@ -81,12 +84,6 @@
 /* Define if sys_siglist[] is in the C library */
 #undef HAVE_SYS_SIGLIST
 
-/* Define if sys_siglist[] are defined in <signal.h> or <unistd.h> */
-#undef SYS_SIGLIST_DECLARED
-
-/* Define if you have a sane <unistd.h> header file */
-#undef HAVE_UNISTD_H
-
 /* Define if you have a sane <termios.h> header file */
 #undef HAVE_TERMIOS_H
 
@@ -99,6 +96,9 @@
 /* Define if you have a bcopy() function in your C library */
 #undef HAVE_BCOPY
 
+/* Define if you have a lstat() function in your C library */
+#undef HAVE_LSTAT
+
 /* Define if you have a sane <termio.h> header file */
 #undef HAVE_TERMIO_H
 
@@ -126,6 +126,45 @@
 /* Define if your OS maps references to /dev/fd/n to file descriptor n */
 #undef HAVE_DEV_FD
 
+/*   ------- ------ ----- ---- --- -- - - -- --- ---- ----- ------ -------   */
+/*   Defines from here on down are enable/disable options to configure */
+
+/* Default PATH (see comments in configure.in for more details) */
+#undef DEFAULT_PATH
+
+/* Include ksh features? (see comments in configure.in for more details) */
+#undef KSH
+
+/* Include emacs editing? (see comments in configure.in for more details) */
+#undef EMACS
+
+/* Include vi editing? (see comments in configure.in for more details) */
+#undef VI
+
+/* Include job control? (see comments in configure.in for more details) */
+#undef JOBS
+
+/* Include brace-expansion? (see comments in configure.in for more details) */
+#undef BRACE_EXPAND
+
+/* Include any history? (see comments in configure.in for more details) */
+#undef HISTORY
+
+/* Include complex history? (see comments in configure.in for more details) */
+#undef COMPLEX_HISTORY
+
+/* Strict POSIX behaviour? (see comments in configure.in for more details) */
+#undef POSIXLY_CORRECT
+
+/* Specify default $ENV? (see comments in configure.in for more details) */
+#undef DEFAULT_ENV
+
+/* Include shl(1) support? (see comments in configure.in for more details) */
+#undef SWTCH
+
+/* Include game-of-life? (see comments in configure.in for more details) */
+#undef SILLY
+
 @BOTTOM@
 
 /* Need to use a separate file to keep the configure script from commenting
