@@ -7,23 +7,33 @@
 #include <ctype.h>
 #include "ksh_stat.h"
 
+/* Does ps4 get parameter substitutions done? */
+#ifdef KSH
+# define PS4_SUBSTITUTE(s)	substitute((s), 0)
+#else
+# define PS4_SUBSTITUTE(s)	(s)
+#endif /* KSH */
+
 static int	comexec	 ARGS((struct op *t, struct tbl *volatile tp, char **ap,
 			      int volatile flags));
 static void	scriptexec ARGS((struct op *tp, char **ap));
 static int	call_builtin ARGS((struct tbl *tp, char **wp));
-static int	iosetup ARGS((struct ioword *iop));
+static int	iosetup ARGS((struct ioword *iop, struct tbl *tp));
 static int	herein ARGS((char *hname, int sub));
 #ifdef KSH
-static char 	*do_selectargs ARGS((char **ap));
-static int	pr_menu ARGS((register char **ap));
+static char 	*do_selectargs ARGS((char **ap, bool_t print_menu));
 #endif /* KSH */
 #ifdef KSH
 static int	dbteste_isa ARGS((Test_env *te, Test_meta meta));
-static char	*dbteste_getopnd ARGS((Test_env *te, Test_op op, int do_eval));
-static int	dbteste_eval ARGS((Test_env *te, Test_op op, char *opnd1,
-				char *opnd2, int do_eval));
-static void	dbteste_error ARGS((Test_env *te, int offset, char *msg));
+static const char *dbteste_getopnd ARGS((Test_env *te, Test_op op,
+					 int do_eval));
+static int	dbteste_eval ARGS((Test_env *te, Test_op op, const char *opnd1,
+				const char *opnd2, int do_eval));
+static void	dbteste_error ARGS((Test_env *te, int offset, const char *msg));
 #endif /* KSH */
+#ifdef OS2
+static int	search_access1 ARGS((const char *path, int mode));
+#endif /* OS2 */
 
 
 /*
@@ -75,6 +85,14 @@ execute(t, flags)
 	if (t == NULL)
 		return 0;
 
+	/* Is this the end of a pipeline?  If so, we want to evaluate the
+	 * command arguments
+	bool_t eval_done = FALSE;
+	if ((flags&XFORK) && !(flags&XEXEC) && (flags&XPCLOSE)) {
+		eval_done = TRUE;
+		tp = eval_execute_args(t, &ap);
+	}
+	 */
 	if ((flags&XFORK) && !(flags&XEXEC) && t->type != TPIPE)
 		return exchild(t, flags, -1); /* run in sub-process */
 
@@ -91,10 +109,10 @@ execute(t, flags)
 		/* POSIX says expand command words first, then redirections,
 		 * and assignments last..
 		 */
-		ap = eval(t->args, t->evalflags | DOBLANK | DOGLOB | DOTILDE);
+		ap = eval(t->args, t->u.evalflags | DOBLANK | DOGLOB | DOTILDE);
 		if (Flag(FXTRACE) && ap[0]) {
 			shf_fprintf(shl_out, "%s",
-				substitute(strval(global("PS4")), 0));
+				PS4_SUBSTITUTE(str_val(global("PS4"))));
 			for (i = 0; ap[i]; i++)
 				shf_fprintf(shl_out, "%s%s", ap[i],
 					ap[i + 1] ? space : newline);
@@ -108,26 +126,19 @@ execute(t, flags)
 		e->savefd = (short *) alloc(sizeofN(short, NUFILE), ATEMP);
 		/* initialize to not redirected */
 		memset(e->savefd, 0, sizeofN(short, NUFILE));
-		/* mark fd 0/1 in-use if pipeline */
-		/* Don't think this is needed any more... -mhr, Nov 10/94
-		if (flags&XPIPEI)
-			e->savefd[0] = -1;
-		if (flags&XPIPEO)
-			e->savefd[1] = -1;
-		*/
 	}
 
 	/* do redirection, to be restored in quitenv() */
 	if (t->ioact != NULL)
 		for (iowp = t->ioact; *iowp != NULL; iowp++) {
-			if (iosetup(*iowp) < 0) {
+			if (iosetup(*iowp, tp) < 0) {
 				exstat = rv = 1;
 				/* Redirection failures for special commands
 				 * cause (non-interactive) shell to exit.
 				 */
 				if (tp && tp->type == CSHELL
 				    && (tp->flag & SPEC_BI))
-					errorf((char *) 0);
+					errorf(null);
 				/* Deal with FERREXIT, quitenv(), etc. */
 				goto Break;
 			}
@@ -145,9 +156,9 @@ execute(t, flags)
 	  case TPIPE:
 		flags |= XFORK;
 		flags &= ~XEXEC;
-		e->savefd[0] = savefd(0);
+		e->savefd[0] = savefd(0, 0);
 		(void) ksh_dup2(e->savefd[0], 0, FALSE); /* stdin of first */
-		e->savefd[1] = savefd(1);
+		e->savefd[1] = savefd(1, 0);
 		while (t->type == TPIPE) {
 			openpipe(pv);
 			(void) ksh_dup2(pv[1], 1, FALSE); /* stdout of curr */
@@ -178,20 +189,43 @@ execute(t, flags)
 		rv = execute(t, flags & XERROK);
 		break;
 
+#ifdef KSH
 	  case TCOPROC:
 	  {
+# ifdef JOB_SIGS
+		sigset_t	omask;
+# endif /* JOB_SIGS */
+
+# ifdef JOB_SIGS
+		/* Block sigchild as we are using things changed in the
+		 * signal handler
+		 */
+		sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
+		e->type = E_ERRH;
+		i = ksh_sigsetjmp(e->jbuf, 0);
+		if (i) {
+			sigprocmask(SIG_SETMASK, &omask, (sigset_t *) 0);
+			quitenv();
+			unwind(i);
+			/*NOTREACHED*/
+		}
+# endif /* JOB_SIGS */
+		/* Already have a (live) co-process? */
 		if (coproc.job && coproc.write >= 0)
 			errorf("coprocess already exists");
+
 		/* Can we re-use the existing co-process pipe? */
-		cleanup_coproc(TRUE);
+		coproc_cleanup(TRUE);
+
 		/* do this before opening pipes, in case these fail */
-		e->savefd[0] = savefd(0);
-		e->savefd[1] = savefd(1);
+		e->savefd[0] = savefd(0, 0);
+		e->savefd[1] = savefd(1, 0);
 
 		openpipe(pv);
 		ksh_dup2(pv[0], 0, FALSE);
 		close(pv[0]);
 		coproc.write = pv[1];
+		coproc.job = (void *) 0;
 
 		if (coproc.readw >= 0)
 			ksh_dup2(coproc.readw, 1, FALSE);
@@ -200,14 +234,25 @@ execute(t, flags)
 			coproc.read = pv[0];
 			ksh_dup2(pv[1], 1, FALSE);
 			coproc.readw = pv[1];	 /* closed before first read */
+			coproc.njobs = 0;
+			/* create new coprocess id */
+			++coproc.id;
 		}
+# ifdef JOB_SIGS
+		sigprocmask(SIG_SETMASK, &omask, (sigset_t *) 0);
+		e->type = E_EXEC; /* no more need for error handler */
+# endif /* JOB_SIGS */
 
-		/* exchild() closes coproc.* in child after fork */
+		/* exchild() closes coproc.* in child after fork,
+		 * will also increment coproc.njobs when the
+		 * job is actually created.
+		 */
 		flags &= ~XEXEC;
 		exchild(t->left, flags|XBGND|XFORK|XCOPROC|XCCLOSE,
 			coproc.readw);
 		break;
 	  }
+#endif /* KSH */
 
 	  case TASYNC:
 		/* XXX non-optimal, I think - "(foo &)", forks for (),
@@ -221,7 +266,7 @@ execute(t, flags)
 	  case TAND:
 		rv = execute(t->left, XERROK);
 		if (t->right != NULL && (rv == 0) == (t->type == TAND))
-			rv = execute(t->right, 0);
+			rv = execute(t->right, flags & XERROK);
 		else
 			flags |= XERROK;
 		break;
@@ -250,12 +295,17 @@ execute(t, flags)
 	  case TFOR:
 #ifdef KSH
 	  case TSELECT:
+	    {
+		volatile bool_t is_first = TRUE;
 #endif /* KSH */
 		ap = (t->vars != NULL) ?
 			  eval(t->vars, DOBLANK|DOGLOB|DOTILDE)
 			: e->loc->argv + 1;
 		e->type = E_LOOP;
-		while ((i = setjmp(e->jbuf)))
+		while (1) {
+			i = ksh_sigsetjmp(e->jbuf, 0);
+			if (!i)
+				break;
 			if ((e->flags&EF_BRKCONT_PASS)
 			    || (i != LBREAK && i != LCONTIN))
 			{
@@ -265,6 +315,7 @@ execute(t, flags)
 				rv = 0;
 				goto Break;
 			}
+		}
 		rv = 0; /* in case of a continue */
 		if (t->type == TFOR) {
 			struct tbl *vq;
@@ -282,10 +333,11 @@ execute(t, flags)
 			struct tbl *vq;
 
 			for (;;) {
-				if ((cp = do_selectargs(ap)) == (char *) 0) {
+				if (!(cp = do_selectargs(ap, is_first))) {
 					rv = 1;
 					break;
 				}
+				is_first = FALSE;
 				vq = global(t->str);
 				if (vq->flag & RDONLY)
 					errorf("%s is read only", t->str);
@@ -293,13 +345,17 @@ execute(t, flags)
 				rv = execute(t->left, flags & XERROK);
 			}
 		}
+	    }
 #endif /* KSH */
 		break;
 
 	  case TWHILE:
 	  case TUNTIL:
 		e->type = E_LOOP;
-		while ((i = setjmp(e->jbuf)))
+		while (1) {
+			i = ksh_sigsetjmp(e->jbuf, 0);
+			if (!i)
+				break;
 			if ((e->flags&EF_BRKCONT_PASS)
 			    || (i != LBREAK && i != LCONTIN))
 			{
@@ -309,6 +365,7 @@ execute(t, flags)
 				rv = 0;
 				goto Break;
 			}
+		}
 		rv = 0; /* in case of a continue */
 		while ((execute(t->left, XERROK) == 0) == (t->type == TWHILE))
 			rv = execute(t->right, flags & XERROK);
@@ -340,7 +397,7 @@ execute(t, flags)
 		break;
 
 	  case TFUNCT:
-		rv = define(t->str, t->left);
+		rv = define(t->str, t);
 		break;
 
 	  case TTIME:
@@ -371,10 +428,8 @@ execute(t, flags)
 	if ((flags&XEXEC))
 		exit(rv);	/* exit child */
 	if (rv != 0 && !(flags & XERROK)) {
-		if (Flag(FERREXIT)) {
-			exstat = rv;
+		if (Flag(FERREXIT))
 			unwind(LERROR);
-		}
 		trapsig(SIGERR_);
 	}
 	return rv;
@@ -395,7 +450,7 @@ comexec(t, tp, ap, flags)
 	int rv = 0;
 	register char *cp;
 	register char **lastp;
-	static struct op texec = {TEXEC};
+	static struct op texec; /* Must be static (XXX but why?) */
 	int type_flags;
 	int keepasn_ok;
 	int fcflags = FC_BI|FC_FUNC|FC_PATH;
@@ -477,15 +532,17 @@ comexec(t, tp, ap, flags)
 			break;
 		tp = findcom(ap[0], fcflags & (FC_BI|FC_FUNC));
 	}
-	/* todo: POSIX says assignments preceding a function are kept, at&t
-	 * ksh does not do this
-	 */
-	if (keepasn_ok && (!ap[0] || (tp && tp->flag & KEEPASN)))
+	if (keepasn_ok && (!ap[0] || (tp && (tp->flag & KEEPASN))))
 		type_flags = 0;
 	else {
 		/* create new variable/function block */
 		newblock();
-		type_flags = LOCAL|EXPORT;
+		/* ksh functions don't keep assignments, POSIX functions do. */
+		if (keepasn_ok && tp && tp->type == CFUNC
+		    && !(tp->flag & FKSH))
+			type_flags = 0;
+		else
+			type_flags = LOCAL|LOCAL_COPY|EXPORT;
 	}
 	if (Flag(FEXPORT))
 		type_flags |= EXPORT;
@@ -494,7 +551,7 @@ comexec(t, tp, ap, flags)
 		if (Flag(FXTRACE)) {
 			if (i == 0)
 				shf_fprintf(shl_out, "%s",
-					substitute(strval(global("PS4")), 0));
+					PS4_SUBSTITUTE(str_val(global("PS4"))));
 			shf_fprintf(shl_out, "%s%s", cp,
 				t->vars[i + 1] ? space : newline);
 			if (!t->vars[i + 1])
@@ -507,7 +564,7 @@ comexec(t, tp, ap, flags)
 		rv = subst_exstat;
 		goto Leave;
 	} else if (!tp) {
-		if (Flag(FRESTRICTED) && strchr_dirsep(cp)) {
+		if (Flag(FRESTRICTED) && ksh_strchr_dirsep(cp)) {
 			warningf(TRUE, "%s: restricted", cp);
 			rv = 1;
 			goto Leave;
@@ -523,24 +580,30 @@ comexec(t, tp, ap, flags)
 	  case CFUNC:			/* function call */
 	  {
 		volatile int old_xflag;
-		volatile int old_inuse;
-		char *volatile old_kshname;
+		volatile Tflag old_inuse;
+		const char *volatile old_kshname;
 
 		if (!(tp->flag & ISSET)) {
 			struct tbl *ftp;
 
 			if (!tp->u.fpath) {
-				/* XXX: exit code 126 vs 127 */
-				warningf(TRUE,
+				if (tp->u2.errno_) {
+					warningf(TRUE,
+				"%s: can't find function definition file - %s",
+						cp, strerror(tp->u2.errno_));
+					rv = 126;
+				} else {
+					warningf(TRUE,
 				"%s: can't find function definition file", cp);
-				rv = 127;
+					rv = 127;
+				}
 				break;
 			}
-			if (include(tp->u.fpath, 0, (char **) 0) < 0) {
+			if (include(tp->u.fpath, 0, (char **) 0, 0) < 0) {
 				warningf(TRUE,
 			    "%s: can't open function definition file %s - %s",
 					cp, tp->u.fpath, strerror(errno));
-				rv = 126;
+				rv = 127;
 				break;
 			}
 			if (!(ftp = findfunc(cp, hash(cp), FALSE))
@@ -555,9 +618,11 @@ comexec(t, tp, ap, flags)
 			tp = ftp;
 		}
 
-		/* posix says $0 remains unchanged, at&t ksh changes it */
+		/* ksh functions set $0 to function name, POSIX functions leave
+		 * $0 unchanged.
+		 */
 		old_kshname = kshname;
-		if (!Flag(FPOSIX))
+		if (tp->flag & FKSH)
 			kshname = ap[0];
 		e->loc->argv = ap;
 		for (i = 0; *ap++ != NULL; i++)
@@ -572,7 +637,8 @@ comexec(t, tp, ap, flags)
 		tp->flag |= FINUSE;
 
 		e->type = E_FUNC;
-		if ((i = setjmp(e->jbuf)) == 0) {
+		i = ksh_sigsetjmp(e->jbuf, 0);
+		if (i == 0) {
 			/* seems odd to pass XERROK here, but at&t ksh does */
 			exstat = execute(tp->val.t, flags & XERROK);
 			i = LRETURN;
@@ -581,7 +647,8 @@ comexec(t, tp, ap, flags)
 		Flag(FXTRACE) = old_xflag;
 		tp->flag = (tp->flag & ~FINUSE) | old_inuse;
 		/* Were we deleted while executing?  If so, free the execution
-		 * tree.  Unfortunately, the table entry is never re-used.
+		 * tree.  todo: Unfortunately, the table entry is never re-used
+		 * until the lookup table is expanded.
 		 */
 		if ((tp->flag & (FDELETE|FINUSE)) == FDELETE) {
 			if (tp->flag & ALLOC) {
@@ -612,38 +679,26 @@ comexec(t, tp, ap, flags)
 	  case CEXEC:		/* executable command */
 	  case CTALIAS:		/* tracked alias */
 		if (!(tp->flag&ISSET)) {
-			/*
-			 * mlj addition:
-			 *
-			 * If you specify a full path to a file
-			 * (or type the name of a file in .) which
-			 * doesn't have execute priv's, it used to
-			 * just say "not found".  Kind of annoying,
-			 * particularly if you just wrote a script
-			 * but forgot to say chmod 755 script.
-			 *
-			 * This should probably be done in eaccess(),
-			 * but it works here (at least I haven't noticed
-			 * changing errno here breaking something
-			 * else).
-			 *
-			 * So, we assume that if the file exists, it
-			 * doesn't have execute privs; else, it really
-			 * is not found.
+			/* errno_ will be set if the named command was found
+			 * but could not be executed (permissions, no execute
+			 * bit, directory, etc).  Print out a (hopefully)
+			 * useful error message and set the exit status to 126.
 			 */
-			if (access(cp, F_OK) < 0)
+			if (tp->u2.errno_) {
+				warningf(TRUE, "%s: cannot execute - %s", cp,
+					strerror(tp->u2.errno_));
+				rv = 126;	/* POSIX */
+			} else {
 				warningf(TRUE, "%s: not found", cp);
-			else
-				warningf(TRUE, "%s: cannot execute", cp);
-			/* XXX posix says 126 if in path and cannot execute */
-			rv = 127;
+				rv = 127;
+			}
 			break;
 		}
 
 		/* set $_ to program's full path */
 		setstr(typeset("_", LOCAL|EXPORT, 0, 0, 0), tp->val.s);
 
-		if ((flags&XEXEC)) {
+		if (flags&XEXEC) {
 			j_exit();
 			if (!(flags&XBGND) || Flag(FMONITOR)) {
 				setexecsig(&sigtraps[SIGINT], SS_RESTORE_ORIG);
@@ -652,6 +707,7 @@ comexec(t, tp, ap, flags)
 		}
 
 		/* to fork we set up a TEXEC node and call execute */
+		texec.type = TEXEC;
 		texec.left = t;	/* for tprint */
 		texec.str = tp->val.s;
 		texec.args = ap;
@@ -673,9 +729,9 @@ scriptexec(tp, ap)
 {
 	char *shell;
 
-	shell = strval(global(EXECSHELL_STR));
+	shell = str_val(global(EXECSHELL_STR));
 	if (shell && *shell)
-		shell = search(shell, path, X_OK);
+		shell = search(shell, path, X_OK, (int *) 0);
 	if (!shell || !*shell)
 		shell = EXECSHELL;
 
@@ -703,19 +759,22 @@ scriptexec(tp, ap)
 				cp++;
 			if (*cp && *cp != '\n') {
 				char *a0 = cp, *a1 = (char *) 0;
+# ifdef OS2
+				char *a2 = cp;
+# endif /* OS2 */
 
 				while (*cp && *cp != '\n' && *cp != ' '
 				       && *cp != '\t')
 				{
-#ifdef OS2
-			/* Allow shell search without prepended path.
+# ifdef OS2
+			/* Allow shell search without prepended path
+			 * if shell with / in pathname cannot be found.
 			 * Use / explicitly so \ can be used if explicit
-			 * path is needed.
-			 * [This could surprise people...]
+			 * needs to be forced.
 			 */
 					if (*cp == '/')
-						a0 = cp + 1;
-#endif /* OS2 */
+						a2 = cp + 1;
+# endif /* OS2 */
 					cp++;
 				}
 				if (*cp && *cp != '\n') {
@@ -734,15 +793,31 @@ scriptexec(tp, ap)
 					*cp = '\0';
 					if (a1)
 						*tp->args-- = a1;
+# ifdef OS2
+					if (a0 != a2 && search_access(a0, X_OK, (int *) 0))
+						a0 = a2;
+# endif /* OS2 */
 					shell = a0;
 				}
 			}
 # ifdef OS2
 		} else {
-			/* Should really check shell type before
-			 * making this assumption.
+		        /* Use ksh documented shell default if present
+			 * else use OS2_SHELL which is assumed to need
+			 * the /c option and '\' as dir separater.
 			 */
-			*tp->args-- = "/c";
+		         char *p = shell;
+
+			 shell = str_val(global("EXECSHELL"));
+			 if (shell && *shell)
+				 shell = search(shell, path, X_OK, (int *) 0);
+			 if (!shell || !*shell) {
+				 shell = p;
+				 *tp->args-- = "/c";
+				 for (p = tp->str; *p; p++)
+					 if (*p == '/')
+						 *p = '\\';
+			 }
 # endif /* OS2 */
 		}
 	}
@@ -773,8 +848,8 @@ shcomexec(wp)
  */
 struct tbl *
 findfunc(name, h, create)
-	char	*name;
-	int	h;
+	const char *name;
+	unsigned int	h;
 	int	create;
 {
 	struct block *l;
@@ -782,12 +857,13 @@ findfunc(name, h, create)
 
 	for (l = e->loc; l; l = l->next) {
 		tp = tsearch(&l->funs, name, h);
-		if (tp && (tp->flag & DEFINED))
+		if (tp)
 			break;
 		if (!l->next && create) {
 			tp = tenter(&l->funs, name, h);
 			tp->flag = DEFINED;
 			tp->type = CFUNC;
+			tp->val.t = (struct op *) 0;
 			break;
 		}
 	}
@@ -795,25 +871,31 @@ findfunc(name, h, create)
 }
 
 /*
- * define function
+ * define function.  Returns 1 if function is being undefined (t == 0) and
+ * function did not exist, returns 0 otherwise.
  */
 int
 define(name, t)
-	char	*name;
+	const char *name;
 	struct op *t;
 {
-	register struct tbl *tp;
+	struct tbl *tp;
+	int was_set = 0;
 
-	tp = findfunc(name, hash(name), TRUE);
-
-	/* If this function is currently being executed, we zap this
-	 * table entry so findfunc() won't see it
-	 */
-	if (tp->flag & FINUSE) {
-		tp->name[0] = '\0';
-		tp->flag &= ~DEFINED; /* ensure it won't be found */
-		tp->flag |= FDELETE;
+	while (1) {
 		tp = findfunc(name, hash(name), TRUE);
+
+		if (tp->flag & ISSET)
+			was_set = 1;
+		/* If this function is currently being executed, we zap this
+		 * table entry so findfunc() won't see it
+		 */
+		if (tp->flag & FINUSE) {
+			tp->name[0] = '\0';
+			tp->flag &= ~DEFINED; /* ensure it won't be found */
+			tp->flag |= FDELETE;
+		} else
+			break;
 	}
 
 	if (tp->flag & ALLOC) {
@@ -823,11 +905,13 @@ define(name, t)
 
 	if (t == NULL) {		/* undefine */
 		tdelete(tp);
-		return 0;
+		return was_set ? 0 : 1;
 	}
 
-	tp->val.t = tcopy(t, tp->areap);
+	tp->val.t = tcopy(t->left, tp->areap);
 	tp->flag |= (ISSET|ALLOC);
+	if (t->u.ksh_func)
+	    tp->flag |= FKSH;
 
 	return 0;
 }
@@ -837,11 +921,11 @@ define(name, t)
  */
 void
 builtin(name, func)
-	char *name;
-	int (*func)();
+	const char *name;
+	int (*func) ARGS((char **));
 {
 	register struct tbl *tp;
-	int flag;
+	Tflag flag;
 
 	/* see if any flags should be set for this builtin */
 	for (flag = 0; ; name++) {
@@ -867,7 +951,7 @@ builtin(name, func)
  */
 struct tbl *
 findcom(name, flags)
-	char	*name;
+	const char *name;
 	int	flags;		/* FC_* */
 {
 	static struct tbl temp;
@@ -877,7 +961,7 @@ findcom(name, flags)
 	char *fpath;			/* for function autoloading */
 	char *npath;
 
-	if (strchr_dirsep(name) != NULL) {
+	if (ksh_strchr_dirsep(name) != NULL) {
 		insert = 0;
 		/* prevent FPATH search below */
 		flags &= ~FC_FUNC;
@@ -892,10 +976,12 @@ findcom(name, flags)
 	if (!tp && (flags & FC_FUNC)) {
 		tp = findfunc(name, h, FALSE);
 		if (tp && !(tp->flag & ISSET)) {
-			if ((fpath = strval(global("FPATH"))) == null)
+			if ((fpath = str_val(global("FPATH"))) == null) {
 				tp->u.fpath = (char *) 0;
-			else
-				tp->u.fpath = search(name, fpath, R_OK);
+				tp->u2.errno_ = 0;
+			} else
+				tp->u.fpath = search(name, fpath, R_OK,
+					&tp->u2.errno_);
 		}
 	}
 	if (!tp && (flags & FC_REGBI) && tbi && (tbi->flag & REG_BI))
@@ -936,13 +1022,14 @@ findcom(name, flags)
 			tp->flag = DEFINED;	/* make ~ISSET */
 		}
 		npath = search(name, flags & FC_DEFPATH ? def_path : path,
-				X_OK);
+				X_OK, &tp->u2.errno_);
 		if (npath) {
-			tp->val.s = tp == &temp ? npath : strsave(npath, APERM);
+			tp->val.s = tp == &temp ? npath : str_save(npath, APERM);
 			tp->flag |= ISSET|ALLOC;
 		} else if ((flags & FC_FUNC)
-			   && (fpath = strval(global("FPATH"))) != null
-			   && (npath = search(name, fpath, R_OK)) != (char *) 0)
+			   && (fpath = str_val(global("FPATH"))) != null
+			   && (npath = search(name, fpath, R_OK,
+					      &tp->u2.errno_)) != (char *) 0)
 		{
 			/* An undocumented feature of at&t ksh is that it
 			 * searches FPATH if a command is not found, even
@@ -965,9 +1052,10 @@ void
 flushcom(all)
 	int all;		/* just relative or all */
 {
-	register struct tbl *tp;
+	struct tbl *tp;
+	struct tstate ts;
 
-	for (twalk(&taliases); (tp = tnext()) != NULL; )
+	for (twalk(&ts, &taliases); (tp = tnext(&ts)) != NULL; )
 		if ((tp->flag&ISSET) && (all || !ISDIRSEP(tp->val.s[0]))) {
 			if (tp->flag&ALLOC) {
 				tp->flag &= ~(ALLOC|ISSET);
@@ -978,35 +1066,47 @@ flushcom(all)
 }
 
 /* Check if path is something we want to find.  Returns -1 for failure. */
-static int
-search_access(path, mode)
-	char *path;
+int
+search_access(path, mode, errnop)
+	const char *path;
 	int mode;
+	int *errnop;		/* set if candidate found, but not suitable */
 {
-	int ret = eaccess(path, mode);
+#ifndef OS2
+	int ret, err = 0;
 	struct stat statb;
 
-	/* if executable pipes come along, this will have to change */
-	if (ret == 0 && mode == X_OK
-	    && (stat(path, &statb) < 0 || !S_ISREG(statb.st_mode)))
+	if (stat(path, &statb) < 0)
+		return -1;
+	ret = eaccess(path, mode);
+	if (ret < 0)
+		err = errno; /* File exists, but we can't access it */
+	else if (mode == X_OK && (!S_ISREG(statb.st_mode)
+		   /* This 'cause access() says root can execute everything */
+		   || !(statb.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH))))
+	{
 		ret = -1;
+		err = S_ISDIR(statb.st_mode) ? EISDIR : EACCES;
+	}
+	if (err && errnop)
+		*errnop = err;
 	return ret;
-}
-
-#ifdef OS2
-static int
-search_access1(path, mode)
-	char *path;
-	int mode;
-{
-	static char *xsuffixes[] = { ".ksh", ".sh", ".cmd", ".exe", ".com",
-				     ".bat", ".", (char *) 0
+#else /* !OS2 */
+	/*
+	 * NOTE: ASSUMES path can be modified and has enough room at the
+	 *       end of the string for a suffix (ie, 4 extra characters).
+	 *	 Certain code knows this (eg, eval.c(globit()),
+	 *	 exec.c(search())).
+	 */
+	static char *xsuffixes[] = { ".ksh", ".exe", ".", ".sh", ".cmd",
+				     ".com", ".bat", (char *) 0
 				   };
-	static char *rsuffixes[] = { ".ksh", ".sh", ".cmd", ".bat", ".",
+	static char *rsuffixes[] = { ".ksh", ".", ".sh", ".cmd", ".bat",
 				      (char *) 0
 				   };
 	int i;
-	char *tp = path + strlen(path);
+	char *mpath = (char *) path;
+	char *tp = mpath + strlen(mpath);
 	char *p;
 	char **sfx;
  
@@ -1015,19 +1115,10 @@ search_access1(path, mode)
 	 * the access test to R_OK...
 	 * This code assumes OS/2 files can have only one suffix...
 	 */
-	if ((p = strrchr((p = strrchr_dirsep(path)) ? p : path, '.'))) {
-		if (mode == X_OK) {
-			for (i = 0; xsuffixes[i]; i++)
-				if (xsuffixes[i][1]
-				    && strcmp(p, xsuffixes[i]) == 0)
-				{
-					mode = R_OK;
-					break;
-				}
-			if (mode != R_OK)
-				return -1;
-		}
-		return search_access(path, mode);
+	if ((p = strrchr((p = ksh_strrchr_dirsep(mpath)) ? p : mpath, '.'))) {
+		if (mode == X_OK) 
+			mode = R_OK;
+		return search_access1(mpath, mode, errnop);
 	}
 	/* Try appending the various suffixes.  Different suffixes for
 	 * read and execute 'cause we don't want to read an executable...
@@ -1035,11 +1126,36 @@ search_access1(path, mode)
 	sfx = mode == R_OK ? rsuffixes : xsuffixes;
 	for (i = 0; sfx[i]; i++) {
 		strcpy(tp, p = sfx[i]);
-		if (search_access(path, R_OK) == 0)
+		if (search_access1(mpath, R_OK, errnop) == 0)
 			return 0;
 		*tp = '\0';
 	}
 	return -1;
+#endif /* !OS2 */
+}
+
+#ifdef OS2
+static int
+search_access1(path, mode, errnop)
+	const char *path;
+	int mode;
+	int *errnop;		/* set if candidate found, but not suitable */
+{
+	int ret, err = 0;
+	struct stat statb;
+
+	if (stat(path, &statb) < 0)
+		return -1;
+	ret = eaccess(path, mode);
+	if (ret < 0)
+		err = errno; /* File exists, but we can't access it */
+	else if (!S_ISREG(statb.st_mode)) {
+		ret = -1;
+		err = S_ISDIR(statb.st_mode) ? EISDIR : EACCES;
+	}
+	if (err && errnop)
+		*errnop = err;
+	return ret;
 }
 #endif /* OS2 */
 
@@ -1047,14 +1163,19 @@ search_access1(path, mode)
  * search for command with PATH
  */
 char *
-search(name, path, mode)
-	char *name, *path;
+search(name, path, mode, errnop)
+	const char *name;
+	const char *path;
 	int mode;		/* R_OK or X_OK */
+	int *errnop;		/* set if candidate found, but not suitable */
 {
-	register char *sp, *p, *xp;
+	const char *sp, *p;
+	char *xp;
 	XString xs;
 	int namelen;
 
+	if (errnop)
+		*errnop = 0;
 #ifdef OS2
 	/* Xinit() allocates 8 additional bytes, so appended suffixes won't
 	 * overflow the memory.
@@ -1063,20 +1184,20 @@ search(name, path, mode)
 	Xinit(xs, xp, namelen, ATEMP);
 	memcpy(Xstring(xs, xp), name, namelen);
 
- 	if (strchr_dirsep(name)) {
-		if (search_access1(Xstring(xs, xp), mode) >= 0)
+ 	if (ksh_strchr_dirsep(name)) {
+		if (search_access(Xstring(xs, xp), mode, errnop) >= 0)
 			return Xstring(xs, xp); /* not Xclose() - see above */
 		Xfree(xs, xp);
 		return NULL;
 	}
 
 	/* Look in current context always. (os2 style) */
-	if (search_access1(Xstring(xs, xp), mode) == 0) 
+	if (search_access(Xstring(xs, xp), mode, errnop) == 0) 
 		return Xstring(xs, xp); /* not Xclose() - xp may be wrong */
 #else /* OS2 */
-	if (strchr_dirsep(name)) {
-		if (search_access(name, mode) == 0)
-			return name;
+	if (ksh_strchr_dirsep(name)) {
+		if (search_access(name, mode, errnop) == 0)
+			return (char *) name;
 		return NULL;
 	}
 
@@ -1098,11 +1219,10 @@ search(name, path, mode)
 		sp = p;
 		XcheckN(xs, xp, namelen);
 		memcpy(xp, name, namelen);
+ 		if (search_access(Xstring(xs, xp), mode, errnop) == 0)
 #ifdef OS2
- 		if (search_access1(Xstring(xs, xp), mode) == 0)
  			return Xstring(xs, xp); /* Not Xclose() - see above */
 #else /* OS2 */
-		if (search_access(Xstring(xs, xp), mode) == 0)
 			return Xclose(xs, xp + namelen);
 #endif /* OS2 */
 		if (*sp++ == '\0')
@@ -1136,34 +1256,31 @@ call_builtin(tp, wp)
  * set up redirection, saving old fd's in e->savefd
  */
 static int
-iosetup(iop)
+iosetup(iop, tp)
 	register struct ioword *iop;
+	struct tbl *tp;
 {
 	register int u = -1;
 	char *cp = iop->name;
-	int do_open, UNINITIALIZED(flags);
+	int iotype = iop->flag & IOTYPE;
+	int do_open = 1, do_close = 0, UNINITIALIZED(flags);
+	struct ioword iotmp;
+	struct stat statb;
 
-	if ((iop->flag&IOTYPE) != IOHERE)
-		cp = evalonestr(cp, DOTILDE|DOGLOB);
+	if (iotype != IOHERE)
+		cp = evalonestr(cp, DOTILDE|(Flag(FTALKING) ? DOGLOB : 0));
 
-	if (Flag(FXTRACE)) {
-		struct ioword iotmp;
+	/* Used for tracing and error messages to print expanded cp */
+	iotmp = *iop;
+	iotmp.name = (iotype == IOHERE) ? (char *) 0 : cp;
+	iotmp.flag |= IONAMEXP;
 
-		iotmp = *iop;
-		/* Print the expanded version */
-		iotmp.name = (iop->flag & IOTYPE) == IOHERE ? (char *) 0 : cp;
+	if (Flag(FXTRACE))
 		shellf("%s%s\n",
-			substitute(strval(global("PS4")), 0),
+			PS4_SUBSTITUTE(str_val(global("PS4"))),
 			snptreef((char *) 0, 32, "%R", &iotmp));
-	}
 
-	/* Do not save if it has already been redirected (i.e. "cat >x >y") */
-	if (e->savefd[iop->unit] == 0)
-		/* c_exec() assumes e->savefd[fd] set for any redirections */
-		e->savefd[iop->unit] = savefd(iop->unit);
-
-	do_open = 1;
-	switch (iop->flag&IOTYPE) {
+	switch (iotype) {
 	  case IOREAD:
 		flags = O_RDONLY;
 		break;
@@ -1174,7 +1291,8 @@ iosetup(iop)
 
 	  case IOWRITE:
 		flags = O_WRONLY | O_CREAT | O_TRUNC;
-		if (Flag(FNOCLOBBER) && !(iop->flag & IOCLOB))
+		if (Flag(FNOCLOBBER) && !(iop->flag & IOCLOB)
+		    && (stat(cp, &statb) < 0 || S_ISREG(statb.st_mode)))
 			flags |= O_EXCL;
 		break;
 
@@ -1184,38 +1302,26 @@ iosetup(iop)
 
 	  case IOHERE:
 		do_open = 0;
-		u = herein(cp, iop->flag&IOEVAL);
+		/* herein() returns -2 if error has been printed */
+		u = herein(cp, iop->flag & IOEVAL);
 		/* cp may have wrong name */
 		break;
 
 	  case IODUP:
 	  {
-		char *emsg;
+		const char *emsg;
 
 		do_open = 0;
-		if (*cp == '-' && !cp[1])
-			close(u = iop->unit);
-		else {
-			int saveu;
-			u = check_fd(cp,
-				   X_OK | ((iop->flag & IORDUP) ? R_OK : W_OK),
-				   &emsg);
-			if (u < 0) {
-				warningf(TRUE, "%d>&%s: %s",
-					iop->unit, cp, emsg);
-				return -1;
-			}
-			u = ksh_dup2(saveu = u, iop->unit, TRUE);
-			if (iop->flag & IORDUP)	/* possible <&p */
-				/* Ensure other side of read pipe is closed
-				 * so EOF can be read properly
-				 */
-				coproc_readw_close(saveu);
-			else			/* possible >&p */
-				/* If co-process input is duped,
-				 * close shell's copy
-				 */
-				coproc_write_close(saveu);
+		if (*cp == '-' && !cp[1]) {
+			u = 1009;	 /* prevent error return below */
+			do_close = 1;
+		} else if ((u = check_fd(cp,
+				X_OK | ((iop->flag & IORDUP) ? R_OK : W_OK),
+				&emsg)) < 0)
+		{
+			warningf(TRUE, "%s: %s",
+				snptreef((char *) 0, 32, "%R", &iotmp), emsg);
+			return -1;
 		}
 		break;
 	  }
@@ -1235,16 +1341,45 @@ iosetup(iop)
 		/* herein() may already have printed message */
 		if (u == -1)
 			warningf(TRUE, "cannot %s %s: %s",
-			       (iop->flag&IOTYPE) == IODUP ? "dup"
-				: ((iop->flag&IOTYPE) == IOREAD
-				    || (iop->flag&IOTYPE) == IOHERE) ? "open"
-				   : "create", cp, strerror(errno));
+			       iotype == IODUP ? "dup"
+				: (iotype == IOREAD || iotype == IOHERE) ?
+				    "open" : "create", cp, strerror(errno));
 		return -1;
 	}
-	if (u != iop->unit) {
-		(void) ksh_dup2(u, iop->unit, FALSE);
-		if (iop->flag != IODUP) /* always true */
+	/* Do not save if it has already been redirected (i.e. "cat >x >y"). */
+	if (e->savefd[iop->unit] == 0)
+		/* c_exec() assumes e->savefd[fd] set for any redirections.
+		 * Ask savefd() not to close iop->unit - allows error messages
+		 * to be seen if iop->unit is 2; also means we can't lose
+		 * the fd (eg, both dup2 below and dup2 in restfd() failing).
+		 */
+		e->savefd[iop->unit] = savefd(iop->unit, 1);
+
+	if (do_close)
+		close(iop->unit);
+	else if (u != iop->unit) {
+		if (ksh_dup2(u, iop->unit, TRUE) < 0) {
+			warningf(TRUE,
+				"could not finish (dup) redirection %s: %s",
+				snptreef((char *) 0, 32, "%R", &iotmp),
+				strerror(errno));
+			if (iotype != IODUP)
+				close(u);
+			return -1;
+		}
+		if (iotype != IODUP)
 			close(u);
+#ifdef KSH
+		/* Touching any co-process fd in an empty exec
+		 * causes the shell to close its copies
+		 */
+		else if (tp && tp->type == CSHELL && tp->val.f == c_exec) {
+			if (iop->flag & IORDUP)	/* possible exec <&p */
+				coproc_read_close(u);
+			else			/* possible exec >&p */
+				coproc_write_close(u);
+		}
+#endif /* KSH */
 	}
 	if (u == 2) /* Clear any write errors */
 		shf_reopen(2, SHF_WR, shl_out);
@@ -1265,7 +1400,7 @@ herein(hname, sub)
 	/* ksh -c 'cat << EOF' can cause this... */
 	if (hname == (char *) 0) {
 		warningf(TRUE, "here document missing");
-		return -2;
+		return -2; /* special to iosetup(): don't print error */
 	}
 	if (sub) {
 		char *cp;
@@ -1279,13 +1414,13 @@ herein(hname, sub)
 		if (shf == NULL)
 			return -1;
 		newenv(E_ERRH);
-		if ((i = setjmp(e->jbuf))) {
+		i = ksh_sigsetjmp(e->jbuf, 0);
+		if (i) {
 			if (shf)
 				shf_close(shf);
 			source = osource;
 			quitenv(); /* after shf_close() due to alloc */
-			/* indicate that error message should not be printed */
-			return -2;
+			return -2; /* special to iosetup(): don't print error */
 		}
 		/* set up yylex input from here file */
 		s = pushs(SFILE, ATEMP);
@@ -1300,9 +1435,8 @@ herein(hname, sub)
 		/* write expanded input to another temp file */
 		h = maketemp(ATEMP);
 		h->next = e->temps; e->temps = h;
-		shf = shf_open(h->name, O_WRONLY|O_CREAT|O_TRUNC, 0666, 0);
-		if (shf == NULL
-		    || (fd = open(h->name, O_RDONLY, 0)) < 0)
+		if (!(shf = h->shf) || (fd = open(h->name, O_RDONLY, 0)) < 0)
+			/* shf closeed by error handler */
 			errorf("%s: %s", h->name, strerror(errno));
 		shf_puts(cp, shf);
 		if (shf_close(shf) == EOF) {
@@ -1329,40 +1463,46 @@ herein(hname, sub)
  *	print the args in column form - assuming that we can
  */
 static char *
-do_selectargs(ap)
+do_selectargs(ap, print_menu)
 	register char **ap;
+	bool_t print_menu;
 {
-	extern int c_read ARGS((char **wp));
-
-	static char *read_args[] = {
+	static const char *const read_args[] = {
 					"read", "-r", "REPLY", (char *) 0
 				    };
 	char *s;
-	int i, UNINITIALIZED(argct);
+	int i, argct;
 
+	for (argct = 0; ap[argct]; argct++)
+		;
 	while (1) {
-		argct = pr_menu(ap);
-		shellf("%s", strval(global("PS3")));
-		if (call_builtin(findcom("read", FC_BI), read_args) != 0) {
-			shellf(newline);
+		/* Menu is printed if
+		 *	- this is the first time around the select loop
+		 *	- the user enters a blank line
+		 *	- the REPLY parameter is empty
+		 */
+		if (print_menu || !*str_val(global("REPLY")))
+			pr_menu(ap);
+		shellf("%s", str_val(global("PS3")));
+		if (call_builtin(findcom("read", FC_BI), (char **) read_args))
 			return (char *) 0;
-		}
-		s = strval(global("REPLY"));
-		while (*s && isspace(*s))
-			s++;
+		s = str_val(global("REPLY"));
 		if (*s) {
 			i = atoi(s);
 			return (i >= 1 && i <= argct) ? ap[i - 1] : null;
 		}
+		print_menu = 1;
 	}
 }
 
 struct select_menu_info {
-	char	**args;
+	char	*const *args;
 	int	arg_width;
 	int	num_width;
 } info;
 
+static char *select_fmt_entry ARGS((void *arg, int i, char *buf, int buflen));
+
 /* format a single select menu item */
 static char *
 select_fmt_entry(arg, i, buf, buflen)
@@ -1381,12 +1521,12 @@ select_fmt_entry(arg, i, buf, buflen)
 /*
  *	print a select style menu
  */
-static int
+int
 pr_menu(ap)
-	register char **ap;
+	char *const *ap;
 {
 	struct select_menu_info smi;
-	char **pp;
+	char *const *pp;
 	int nwidth, dwidth;
 	int i, n;
 
@@ -1475,7 +1615,7 @@ dbteste_isa(te, meta)
 	return ret;
 }
 
-static char *
+static const char *
 dbteste_getopnd(te, op, do_eval)
 	Test_env *te;
 	Test_op op;
@@ -1491,7 +1631,7 @@ dbteste_getopnd(te, op, do_eval)
 	if (!do_eval)
 		return null;
 
-	if (op == TO_STGT || op == TO_STLT)
+	if (op == TO_STEQL || op == TO_STNEQ)
 		s = evalstr(s, DOTILDE | DOPAT);
 	else
 		s = evalstr(s, DOTILDE);
@@ -1503,8 +1643,8 @@ static int
 dbteste_eval(te, op, opnd1, opnd2, do_eval)
 	Test_env *te;
 	Test_op op;
-	char *opnd1;
-	char *opnd2;
+	const char *opnd1;
+	const char *opnd2;
 	int do_eval;
 {
 	return test_eval(te, op, opnd1, opnd2, do_eval);
@@ -1514,7 +1654,7 @@ static void
 dbteste_error(te, offset, msg)
 	Test_env *te;
 	int offset;
-	char *msg;
+	const char *msg;
 {
 	te->flags |= TEF_ERROR;
 	internal_errorf(0, "dbteste_error: %s (offset %d)", msg, offset);
