@@ -14,6 +14,7 @@
 # undef S_ISCHR
 # undef S_ISBLK
 # undef S_ISFIFO
+# undef S_ISSOCK
 # undef S_ISLNK
 #endif /* STAT_MACROS_BROKEN */
 
@@ -45,3 +46,13 @@
 #ifndef S_ISVTX
 # define S_ISVTX	01000	/* sticky bit */
 #endif /* S_ISVTX */
+
+#ifndef S_IXUSR
+# define S_IXUSR	00100	/* user execute bit */
+#endif /* S_IXUSR */
+#ifndef S_IXGRP
+# define S_IXGRP	00010	/* user execute bit */
+#endif /* S_IXGRP */
+#ifndef S_IXOTH
+# define S_IXOTH	00001	/* user execute bit */
+#endif /* S_IXOTH */
