@@ -13,6 +13,7 @@ void 	afree		ARGS((void *ptr, Area *ap));
 /* c_ksh.c */
 int 	c_hash		ARGS((char **wp));
 int 	c_cd		ARGS((char **wp));
+int 	c_pwd		ARGS((char **wp));
 int 	c_print		ARGS((char **wp));
 int 	c_whence	ARGS((char **wp));
 int 	c_command	ARGS((char **wp));
@@ -46,38 +47,34 @@ int 	c_exec		ARGS((char **wp));
 int 	c_builtin	ARGS((char **wp));
 /* c_test.c */
 int 	c_test		ARGS((char **wp));
-int	is_db_unop	ARGS((char *s));
-int	is_db_binop	ARGS((char *s));
-int	is_db_patop	ARGS((int op));
-/* edit.c */
+/* edit.c: most prototypes in edit.h */
 void 	x_init		ARGS((void));
 int 	x_read		ARGS((char *buf, size_t len));
-int 	x_getc		ARGS((void));
-void 	x_flush		ARGS((void));
-void 	x_putc		ARGS((int c));
-void 	x_puts		ARGS((char *s));
-bool_t 	x_mode		ARGS((bool_t onoff));
-int 	promptlen	ARGS((char *cp));
-void	set_editmode	ARGS((char *ed));
+void	set_editmode	ARGS((const char *ed));
 /* emacs.c: most prototypes in edit.h */
-int 	x_bind		ARGS((char *a1, char *a2, int macro, int list));
+int 	x_bind		ARGS((const char *a1, const char *a2, int macro,
+			      int list));
 /* eval.c */
 char *	substitute	ARGS((const char *cp, int f));
 char **	eval		ARGS((char **ap, int f));
 char *	evalstr		ARGS((char *cp, int f));
 char *	evalonestr	ARGS((char *cp, int f));
-/* XPtrV not defined at this point... */
-void	expand		ARGS((/*char *cp, XPtrV *wp, int f*/));
+char	*debunk		ARGS((char *dp, const char *sp));
+void	expand		ARGS((char *cp, XPtrV *wp, int f));
+int glob_str		ARGS((char *cp, XPtrV *wp, int markdirs));
 /* exec.c */
 int	fd_clexec	ARGS((int fd));
 int 	execute		ARGS((struct op * volatile t, volatile int flags));
 int 	shcomexec	ARGS((char **wp));
-struct tbl * findfunc	ARGS((char *name, int h, int create));
-int 	define		ARGS((char *name, struct op *t));
-void 	builtin		ARGS((char *name, int (*func)()));
-struct tbl *	findcom	ARGS((char *name, int flags));
+struct tbl * findfunc	ARGS((const char *name, unsigned int h, int create));
+int 	define		ARGS((const char *name, struct op *t));
+void 	builtin		ARGS((const char *name, int (*func)(char **)));
+struct tbl *	findcom	ARGS((const char *name, int flags));
 void 	flushcom	ARGS((int all));
-char *	search		ARGS((char *name, char *path, int mode));
+char *	search		ARGS((const char *name, const char *path, int mode,
+			      int *errnop));
+int	search_access	ARGS((const char *path, int mode, int *errnop));
+int	pr_menu		ARGS((char *const *ap));
 /* expr.c */
 int 	evaluate	ARGS((const char *expr, long *rval, int error_ok));
 int	v_evaluate	ARGS((struct tbl *vp, const char *expr, volatile int error_ok));
@@ -85,18 +82,19 @@ int	v_evaluate	ARGS((struct tbl *vp, const char *expr, volatile int error_ok));
 void	init_histvec	ARGS((void));
 void 	hist_init	ARGS((Source *s));
 void 	hist_finish	ARGS((void));
-void	histsave	ARGS((int lno, char *cmd, int dowrite));
+void	histsave	ARGS((int lno, const char *cmd, int dowrite));
 #ifdef HISTORY
 int 	c_fc	 	ARGS((register char **wp));
 void	sethistsize	ARGS((int n));
-void	sethistfile	ARGS((char *name));
+void	sethistfile	ARGS((const char *name));
 # ifdef EASY_HISTORY
-void 	histappend	ARGS((char *cmd, int nl_seperate));
+void 	histappend	ARGS((const char *cmd, int nl_seperate));
 # endif
 char **	histpos	 	ARGS((void));
 int 	histN	 	ARGS((void));
 int 	histnum	 	ARGS((int n));
-int	findhist	ARGS((int start, int fwd, char *str, int anchored));
+int	findhist	ARGS((int start, int fwd, const char *str,
+			      int anchored));
 #endif /* HISTORY */
 /* io.c */
 void 	errorf		ARGS((const char *fmt, ...))
@@ -115,17 +113,19 @@ void 	shprintf	ARGS((const char *fmt, ...))
 int	can_seek	ARGS((int fd));
 void	initio		ARGS((void));
 int	ksh_dup2	ARGS((int ofd, int nfd, int errok));
-int 	savefd		ARGS((int fd));
+int 	savefd		ARGS((int fd, int noclose));
 void 	restfd		ARGS((int fd, int ofd));
 void 	openpipe	ARGS((int *pv));
 void 	closepipe	ARGS((int *pv));
-int	check_fd	ARGS((char *name, int mode, char **emsgp));
+int	check_fd	ARGS((char *name, int mode, const char **emsgp));
+#ifdef KSH
 void	coproc_init	ARGS((void));
 void	coproc_read_close ARGS((int fd));
 void	coproc_readw_close ARGS((int fd));
 void	coproc_write_close ARGS((int fd));
-int	get_coproc_fd	ARGS((int mode, char **emsgp));
-void	cleanup_coproc	ARGS((int reuse));
+int	coproc_getfd	ARGS((int mode, const char **emsgp));
+void	coproc_cleanup	ARGS((int reuse));
+#endif /* KSH */
 struct temp *maketemp	ARGS((Area *ap));
 /* jobs.c */
 void 	j_init		ARGS((int mflagset));
@@ -134,10 +134,10 @@ void 	j_change	ARGS((void));
 int 	exchild		ARGS((struct op *t, int flags, int close_fd));
 void 	startlast	ARGS((void));
 int 	waitlast	ARGS((void));
-int 	waitfor		ARGS((char *cp, int *sigp));
-int 	j_kill		ARGS((char *cp, int sig));
-int 	j_resume	ARGS((char *cp, int bg));
-int 	j_jobs		ARGS((char *cp, int slp, int nflag));
+int 	waitfor		ARGS((const char *cp, int *sigp));
+int 	j_kill		ARGS((const char *cp, int sig));
+int 	j_resume	ARGS((const char *cp, int bg));
+int 	j_jobs		ARGS((const char *cp, int slp, int nflag));
 void 	j_notify	ARGS((void));
 pid_t	j_async		ARGS((void));
 int 	j_stopped_running	ARGS((void));
@@ -147,17 +147,18 @@ void 	yyerror		ARGS((const char *fmt, ...))
 				GCC_FUNC_ATTR2(noreturn, format(printf, 1, 2));
 Source * pushs		ARGS((int type, Area *areap));
 void	set_prompt	ARGS((int to, Source *s));
-void 	pprompt		ARGS((char *cp, int ntruncate));
+void 	pprompt		ARGS((const char *cp, int ntruncate));
 /* mail.c */
+#ifdef KSH
 void 	mcheck		ARGS((void));
 void 	mbset		ARGS((char *p));
 void 	mpset		ARGS((char *mptoparse));
-void 	mprint		ARGS((void));
+#endif /* KSH */
 /* main.c */
-int 	main		ARGS((int argc, char **argv, char **envp));
-int 	include		ARGS((char *name, int argc, char **argv));
-int 	command		ARGS((char *comm));
-int 	shell		ARGS((Source *volatile s, int volatile exit_atend));
+int 	include		ARGS((const char *name, int argc, char **argv,
+			      int intr_ok));
+int 	command		ARGS((const char *comm));
+int 	shell		ARGS((Source *volatile s, int volatile toplevel));
 void 	unwind		ARGS((int i)) GCC_FUNC_ATTR(noreturn);
 void 	newenv		ARGS((int type));
 void 	quitenv		ARGS((void));
@@ -168,34 +169,37 @@ void 	aerror		ARGS((Area *ap, const char *msg))
 void 	setctypes	ARGS((const char *s, int t));
 void 	initctypes	ARGS((void));
 char *	ulton		ARGS((unsigned long n, int base));
-char *	strsave		ARGS((const char *s, Area *ap));
-char *	strnsave	ARGS((const char *s, int n, Area *ap));
+char *	str_save	ARGS((const char *s, Area *ap));
+char *	str_nsave	ARGS((const char *s, int n, Area *ap));
 int	option		ARGS((const char *n));
 char *	getoptions	ARGS((void));
 void	change_flag	ARGS((enum sh_flag f, int what, int newval));
 int	parse_args	ARGS((char **argv, int what, int *setargsp));
-int 	getn		ARGS((char *as, int *ai));
-int 	bi_getn		ARGS((char *as, int *ai));
+int 	getn		ARGS((const char *as, int *ai));
+int 	bi_getn		ARGS((const char *as, int *ai));
 char *	strerror	ARGS((int i));
-int 	gmatch		ARGS((char *s, char *p, int isfile));
-void 	qsortp		ARGS((void **base, size_t n, int (*f)()));
+int 	gmatch		ARGS((const char *s, const char *p, int isfile));
+int	has_globbing	ARGS((const char *xp, const char *xpe));
+const unsigned char *pat_scan ARGS((const unsigned char *p,
+				const unsigned char *pe, int match_sep));
+void 	qsortp		ARGS((void **base, size_t n, int (*f)(void *, void *)));
 int 	xstrcmp		ARGS((void *p1, void *p2));
 void	ksh_getopt_reset ARGS((Getopt *go, int));
-int	ksh_getopt	ARGS((char **argv, Getopt *go, char *options));
-void	print_value_quoted ARGS((char *s));
+int	ksh_getopt	ARGS((char **argv, Getopt *go, const char *options));
+void	print_value_quoted ARGS((const char *s));
 void	print_columns	ARGS((struct shf *shf, int n,
-			      char *(*func) ARGS((void *, int, char *, int)),
+			      char *(*func)(void *, int, char *, int),
 			      void *arg, int max_width));
 int	strip_nuls	ARGS((char *buf, int nbytes));
-char	*str_zcpy	ARGS((char *dst, char *src, int dsize));
+char	*str_zcpy	ARGS((char *dst, const char *src, int dsize));
 int	blocking_read	ARGS((int fd, char *buf, int nbytes));
 int	reset_nonblock	ARGS((int fd));
 char	*ksh_get_wd	ARGS((char *buf, int bsize));
 /* path.c */
-int	make_path	ARGS((char *cwd, char *file, char **pathlist,
-				XString *xsp, int *phys_pathp));
+int	make_path	ARGS((const char *cwd, const char *file,
+			      char **pathlist, XString *xsp, int *phys_pathp));
 void	simplify_path	ARGS((char *path));
-char	*get_phys_path	ARGS((char *path));
+char	*get_phys_path	ARGS((const char *path));
 void	set_current_wd	ARGS((char *path));
 /* syn.c */
 void 	initkeywords	ARGS((void));
@@ -206,14 +210,16 @@ void 	tinit		ARGS((struct table *tp, Area *ap, int tsize));
 struct tbl *	tsearch	ARGS((struct table *tp, const char *n, unsigned int h));
 struct tbl *	tenter	ARGS((struct table *tp, const char *n, unsigned int h));
 void 	tdelete		ARGS((struct tbl *p));
-void 	twalk		ARGS((struct table *tp));
-struct tbl *	tnext	ARGS((void));
+void 	twalk		ARGS((struct tstate *ts, struct table *tp));
+struct tbl *	tnext	ARGS((struct tstate *ts));
 struct tbl **	tsort	ARGS((struct table *tp));
 /* trace.c */
 /* trap.c */
 void	inittraps	ARGS((void));
+#ifdef KSH
 void	alarm_init	ARGS((void));
-Trap *	gettrap		ARGS((char *name));
+#endif /* KSH */
+Trap *	gettrap		ARGS((const char *name));
 RETSIGTYPE trapsig	ARGS((int i));
 void	intrcheck	ARGS((void));
 int	fatal_trap_check ARGS((void));
@@ -225,36 +231,59 @@ void 	restoresigs	ARGS((void));
 void	settrap		ARGS((Trap *p, char *s));
 int	block_pipe	ARGS((void));
 void	restore_pipe	ARGS((int restore_dfl));
-int	setsig		ARGS((Trap *p, RETSIGTYPE (*f)(), int flags));
+int	setsig		ARGS((Trap *p, handler_t f, int flags));
 void	setexecsig	ARGS((Trap *p, int restore));
 /* tree.c */
-int 	fptreef		ARGS((struct shf *f, int indent, char *fmt, ...));
-char *	snptreef	ARGS((char *s, int n, char *fmt, ...));
+int 	fptreef		ARGS((struct shf *f, int indent, const char *fmt, ...));
+char *	snptreef	ARGS((char *s, int n, const char *fmt, ...));
 struct op *	tcopy	ARGS((struct op *t, Area *ap));
 char *	wdcopy		ARGS((const char *wp, Area *ap));
-const char *wdscan	ARGS((const char *wp, int c));
+char *	wdscan		ARGS((const char *wp, int c));
 void 	tfree		ARGS((struct op *t, Area *ap));
 /* var.c */
 void 	newblock	ARGS((void));
 void 	popblock	ARGS((void));
 void	initvar		ARGS((void));
 struct tbl *	global	ARGS((const char *n));
-struct tbl *	local	ARGS((const char *n));
-char *	strval		ARGS((struct tbl *vp));
+struct tbl *	local	ARGS((const char *n, bool_t copy));
+char *	str_val		ARGS((struct tbl *vp));
 long 	intval		ARGS((struct tbl *vp));
-void 	setstr		ARGS((struct tbl *vq, char *s));
-struct tbl *	strint	ARGS((struct tbl *vq, struct tbl *vp));
+void 	setstr		ARGS((struct tbl *vq, const char *s));
+struct tbl *setint_v	ARGS((struct tbl *vq, struct tbl *vp));
 void 	setint		ARGS((struct tbl *vq, long n));
 int	getint		ARGS((struct tbl *vp, long *nump));
-struct tbl *	typeset	ARGS((char *var, int set, int clr, int field, int base));
+struct tbl *	typeset	ARGS((const char *var, Tflag set, Tflag clr, int field, int base));
 void 	unset		ARGS((struct tbl *vp, int array_ref));
 char  * skip_varname	ARGS((const char *s, int aok));
-char	*skip_wdvarname ARGS((char *s, int aok));
-int	is_wdvarname	ARGS((char *s, int aok));
-int	is_wdvarassign	ARGS((char *s));
+char	*skip_wdvarname ARGS((const char *s, int aok));
+int	is_wdvarname	ARGS((const char *s, int aok));
+int	is_wdvarassign	ARGS((const char *s));
 char **	makenv		ARGS((void));
 int	array_ref_len	ARGS((const char *cp));
-char *  arrayname	ARGS((char *str));
-void    set_array	ARGS((char *var, int reset, char **vals));
+char *	arrayname	ARGS((const char *str));
+void    set_array	ARGS((const char *var, int reset, char **vals));
 /* version.c */
-/* vi.c: prototypes in edit.h */
+/* vi.c: see edit.h */
+
+
+/* Hack to avoid billions of compile warnings on SunOS 4.1.x */
+#if defined(MUN) && defined(sun) && !defined(__svr4__)
+extern void bcopy ARGS((const void *src, void *dst, size_t size));
+extern int fclose ARGS((FILE *fp));
+extern int fprintf ARGS((FILE *fp, const char *fmt, ...));
+extern int fread ARGS((void *buf, int size, int num, FILE *fp));
+extern int ioctl ARGS((int fd, int request, void *arg));
+extern int killpg ARGS((int pgrp, int sig));
+extern int nice ARGS((int n));
+extern int readlink ARGS((const char *path, char *buf, int bufsize));
+extern int setpgrp ARGS((int pid, int pgrp));
+extern int strcasecmp ARGS((const char *s1, const char *s2));
+extern int tolower ARGS((int));
+extern int toupper ARGS((int));
+/*  Include files aren't included yet */
+extern int getrlimit ARGS(( /* int resource, struct rlimit *rpl */ ));
+extern int getrusage ARGS(( /* int who, struct rusage *rusage */ ));
+extern int gettimeofday ARGS(( /* struct timeval *tv, struct timezone *tz */ ));
+extern int setrlimit ARGS(( /* int resource, struct rlimit *rlp */ ));
+extern int lstat ARGS(( /* const char *path, struct stat *buf */ ));
+#endif
