@@ -11,7 +11,7 @@ struct table {
 };
 
 struct tbl {			/* table item */
-	INT32	flag;		/* flags */
+	Tflag	flag;		/* flags */
 	int	type;		/* command type (see below), base (if INTEGER),
 				 * or offset from val.s of value (if EXPORT) */
 	Area	*areap;		/* area to allocate from */
@@ -22,7 +22,10 @@ struct tbl {			/* table item */
 		struct op *t;	/* "function" tree */
 	} val;			/* value */
 	int	index;		/* index for an array */
-	int	field;		/* field with for -L/-R/-Z */
+	union {
+	    int	field;		/* field with for -L/-R/-Z */
+	    int errno_;		/* CEXEC/CTALIAS */
+	} u2;
 	union {
 		struct tbl *array;	/* array values */
 		char *fpath;		/* temporary path to undef function */
@@ -31,32 +34,36 @@ struct tbl {			/* table item */
 };
 
 /* common flag bits */
-#define	ALLOC	BIT(0)		/* val.s has been allocated */
-#define	DEFINED	BIT(1)		/* is defined in block */
-#define	ISSET	BIT(2)		/* has value, vp->val.[si] */
-#define	EXPORT	BIT(3)		/* exported variable/function */
-#define	TRACE	BIT(4)		/* var: user flagged, func: execution tracing */
+#define	ALLOC		BIT(0)	/* val.s has been allocated */
+#define	DEFINED		BIT(1)	/* is defined in block */
+#define	ISSET		BIT(2)	/* has value, vp->val.[si] */
+#define	EXPORT		BIT(3)	/* exported variable/function */
+#define	TRACE		BIT(4)	/* var: user flagged, func: execution tracing */
 /* (start non-common flags at 8) */
 /* flag bits used for variables */
-#define	SPECIAL	BIT(8)		/* PATH, IFS, SECONDS, etc */
-#define	INTEGER	BIT(9)		/* val.i contains integer value */
-#define	RDONLY	BIT(10)		/* read-only variable */
-#define	LOCAL	BIT(11)		/* for local typeset() */
-#define ARRAY	BIT(13)		/* array */
-#define LJUST	BIT(14)		/* left justify */
-#define RJUST	BIT(15)		/* right justify */
-#define ZEROFIL	BIT(16)		/* 0 filled if RJUSTIFY, strip 0s if LJUSTIFY */
-#define LCASEV	BIT(17)		/* convert to lower case */
-#define UCASEV_AL BIT(18)	/* convert to upper case / autoload function */
-#define INT_U	BIT(19)		/* unsigned integer */
-#define INT_L	BIT(20)		/* long integer (no-op) */
-#define IMPORT	BIT(21)		/* flag to typeset(): no arrays, must have = */
+#define	SPECIAL		BIT(8)	/* PATH, IFS, SECONDS, etc */
+#define	INTEGER		BIT(9)	/* val.i contains integer value */
+#define	RDONLY		BIT(10)	/* read-only variable */
+#define	LOCAL		BIT(11)	/* for local typeset() */
+#define ARRAY		BIT(13)	/* array */
+#define LJUST		BIT(14)	/* left justify */
+#define RJUST		BIT(15)	/* right justify */
+#define ZEROFIL		BIT(16)	/* 0 filled if RJUSTIFY, strip 0s if LJUSTIFY */
+#define LCASEV		BIT(17)	/* convert to lower case */
+#define UCASEV_AL	BIT(18)/* convert to upper case / autoload function */
+#define INT_U		BIT(19)	/* unsigned integer */
+#define INT_L		BIT(20)	/* long integer (no-op) */
+#define IMPORT		BIT(21)	/* flag to typeset(): no arrays, must have = */
+#define LOCAL_COPY	BIT(22)	/* with LOCAL - copy attrs from existing var */
+#define EXPRINEVAL	BIT(23)	/* contents currently being evaluated */
+#define EXPRLVALUE	BIT(24)	/* useable as lvalue (temp flag) */
 /* flag bits used for taliases/builtins/aliases/keywords/functions */
-#define KEEPASN	BIT(8)		/* keep command assignments (eg, var=x cmd) */
-#define FINUSE	BIT(9)		/* function being executed */
-#define FDELETE	BIT(10)		/* function deleted while it was executing */
-#define SPEC_BI	BIT(11)		/* a POSIX special builtin */
-#define REG_BI	BIT(12)		/* a POSIX regular builtin */
+#define KEEPASN		BIT(8)	/* keep command assignments (eg, var=x cmd) */
+#define FINUSE		BIT(9)	/* function being executed */
+#define FDELETE		BIT(10)	/* function deleted while it was executing */
+#define FKSH		BIT(11)	/* function defined with function x (vs x()) */
+#define SPEC_BI		BIT(12)	/* a POSIX special builtin */
+#define REG_BI		BIT(13)	/* a POSIX regular builtin */
 
 /* command types */
 #define	CNONE	0		/* undefined */
@@ -76,13 +83,28 @@ struct tbl {			/* table item */
 #define FC_PATH		BIT(4)	/* do path search */
 #define FC_DEFPATH	BIT(5)	/* use default path in path search */
 
+
+#define AF_ARGV_ALLOC	0x1	/* argv[] array allocated */
+#define AF_ARGS_ALLOCED	0x2	/* argument strings allocated */
+#define AI_ARGV(a, i)	((i) == 0 ? (a).argv[0] : (a).argv[(i) - (a).skip])
+#define AI_ARGC(a)	((a).argc_ - (a).skip)
+
+/* Argument info.  Used for $#, $* for shell, functions, includes, etc. */
+struct arg_info {
+	int flags;	/* AF_* */
+	char **argv;
+	int argc_;
+	int skip;	/* first arg is argv[0], second is argv[1 + skip] */
+};
+
 /*
  * activation record for function blocks
  */
 struct block {
 	Area	area;		/* area to allocate things */
-	int	argc;		/* current $# */
-	char **	argv;		/* current $* */
+	/*struct arg_info argi;*/
+	char	**argv;
+	int	argc;
 	struct	table vars;	/* local variables */
 	struct	table funs;	/* local functions */
 #if 1
@@ -94,6 +116,15 @@ struct block {
 	struct	block *next;	/* enclosing block */
 };
 
+/*
+ * Used by twalk() and tnext() routines.
+ */
+struct tstate {
+	int left;
+	struct tbl **next;
+};
+
+
 EXTERN	struct table taliases;	/* tracked aliases */
 EXTERN	struct table builtins;	/* built-in commands */
 EXTERN	struct table aliases;	/* aliases */
@@ -101,8 +132,8 @@ EXTERN	struct table keywords;	/* keywords */
 EXTERN	struct table homedirs;	/* homedir() cache */
 
 struct builtin {
-	char   *name;
-	int  (*func)();
+	const char   *name;
+	int  (*func) ARGS((char **));
 };
 
 /* these really are externs! Look in table.c for them */
@@ -131,8 +162,8 @@ extern const struct builtin shbuiltins [], kshbuiltins [];
 #define PS1	0		/* command */
 #define PS2	1		/* command continuation */
 
-EXTERN	char   *path;		/* PATH value */
-EXTERN	char   *def_path;	/* path to use if PATH not set */
+EXTERN	const char   *path;	/* PATH value */
+EXTERN	const char   *def_path;	/* path to use if PATH not set */
 EXTERN	char   *tmpdir;		/* TMPDIR value */
-EXTERN	char   *prompt;	
+EXTERN	const char   *prompt;
 EXTERN	int	cur_prompt;	/* PS1 or PS2 */
